local trans
local enabledID
local redoID
local finishID
local version = "1.00"
local controls = system.getDeviceType() == "JETI DC-24" and 10 or 4
local sequencers = 4
local sequencer = 1
local startTime = {}
local startIndex = {}
local back = {}
local enabled = {}
local label = {}
local switch = {}
local count = {}
local values = {}
local times = {}
local type = {}
local redo = {}
local finish = {}
local ctrl = {}
local busy = {}

--------------------------------------------------------------------------------
local function readFile(path) 
	local f = io.open(path,"r")
	local lines={}
	if (f) then
		while 1 do 
			local buf=io.read(f,512)
			if (buf ~= "")then 
				lines[#lines+1] = buf
				else
				break   
			end   
		end 
		io.close(f)
		return table.concat(lines,"") 
	end
end 
--------------------------------------------------------------------------------
local function setLanguage()	
	local lng = system.getLocale();
	local file = readFile("Apps/Sequencer/locale.jsn")
	local obj = json.decode(file)  
	if (obj) then
		trans = obj[lng] or obj[obj.default]
	end
end

local function resetSequence(i)
	startTime[i] = nil
	startIndex[i] = nil
	back[i] = nil
end

local function getSequencer(controlNo)
	for i=1, sequencers do
		if (ctrl[i] == controlNo) then
			return i
		end
	end
end

local function registerControl(sequencer)
	for i=1, controls do
		if (enabled[sequencer] and not getSequencer(i)) then
			ctrl[sequencer] = system.registerControl(i, label[i], "AQ"..sequencer, "") 
			if (ctrl[sequencer]) then
				system.pSave("ctrl"..sequencer, ctrl[sequencer])
				return
			end
		end
	end
	
	error("unable to register control")
end

--------------------------------------------------------------------------------
local function initForm(formID)
	sequencer = formID

	for i=1, sequencers do
		form.setButton(i, "AQ"..i, sequencer == i and HIGHLIGHTED or ENABLED)
	end

	form.addRow(1)
	form.addLabel({label=trans.title,font=FONT_BOLD})
	
	form.addRow(2)
	form.addLabel({label=trans.enabled, width=270})
	enabledID = form.addCheckbox(enabled[sequencer], function(value)
		enabled[sequencer] = not value
		system.pSave("enabled"..sequencer, enabled[sequencer] and 1 or 0)
		form.setValue(enabledID, enabled[sequencer])
		resetSequence(sequencer)
		
		if (ctrl[sequencer]) then
			system.unregisterControl(ctrl[sequencer])
			ctrl[sequencer] = nil
		end
		
		if (enabled[sequencer]) then
			registerControl(sequencer)
		end
	end)
	
	form.addRow(2)
	form.addLabel({label=trans.switch})
	form.addInputbox(switch[sequencer],true, function(value)
		switch[sequencer] = value
		system.pSave("switch"..sequencer, value) 
		resetSequence(sequencer)
	end) 
	
	form.addRow(2)
	form.addLabel({label=trans.label})
	form.addTextbox(label[sequencer], 24, function(value)
		label[sequencer] = value
		system.pSave("label"..sequencer, value) 
		resetSequence(sequencer)
		
		if (ctrl[sequencer]) then
			system.unregisterControl(ctrl[sequencer])
			ctrl[sequencer] = nil
		end
		
		if (enabled[sequencer]) then
			registerControl(sequencer)
		end
	end)
	
	form.addRow(2)
	form.addLabel({label=trans.type})
	form.addSelectbox({trans.symmetric, trans.asymmetric}, type[sequencer], false, function(value)
		type[sequencer] = value
		system.pSave("type"..sequencer, value)
		resetSequence(sequencer)
	end)
	
	form.addRow(2)
	form.addLabel({label=trans.loop, width=270})
	redoID = form.addCheckbox(redo[sequencer], function(value)
		redo[sequencer] = not value
		system.pSave("redo"..sequencer, redo[sequencer] and 1 or 0)
		form.setValue(redoID, redo[sequencer])
		resetSequence(sequencer)
	end)
	
	form.addRow(2)
	form.addLabel({label=trans.finish, width=270})
	finishID = form.addCheckbox(finish[sequencer], function(value)
		finish[sequencer] = not value
		system.pSave("finish"..sequencer, finish[sequencer] and 1 or 0)
		form.setValue(finishID, finish[sequencer])
		resetSequence(sequencer)
	end)
	
	form.addRow(2)
	form.addLabel({label=trans.count})
	form.addIntbox(count[sequencer],2,32,2,0,1,function(value)
		count[sequencer] = value
		
		for i=1, 32 do
			values[sequencer][i] = i <= count[sequencer] and (values[sequencer][i] or -100) or nil
			times[sequencer][i] = i < count[sequencer] and (times[sequencer][i] or 50) or nil
		end
		
		system.pSave("count"..sequencer, count[sequencer])
		system.pSave("values"..sequencer, values[sequencer])
		system.pSave("times"..sequencer, times[sequencer])
		resetSequence(sequencer)
		form.reinit(sequencer)
	end)
	
	for i=1, count[sequencer] do
		form.addRow(1)
		form.addLabel({label=trans.position.." "..tostring(i),font=FONT_BOLD})
		
		form.addRow(4)
		form.addLabel({label=trans.value.." "..tostring(i)})
		form.addIntbox(values[sequencer][i] or i ==  2 and 100 or -100,-100,100,-100,0,1,function(value)
			values[sequencer][i] = value
			system.pSave("values"..sequencer, values[sequencer])
			resetSequence(sequencer)
		end)
		
		if (i < count[sequencer]) then
			form.addLabel({label=trans.time.." "..tostring(i).."->"..tostring(i+1)})
			form.addIntbox(times[sequencer][i] or 50,1,300,50,0,1,function(value)
				times[sequencer][i] = value
				system.pSave("times"..sequencer, times[sequencer])
				resetSequence(sequencer)
			end)
		else
			form.addSpacer(1, 1)
			form.addSpacer(1, 1)
		end
	end
	
	form.addRow(1)
	form.addLabel({label="Powered by Thorn - v."..version,font=FONT_MINI, alignRight=true})
end

local function keyPressed(key)
	if (key == KEY_1 and formID ~= 1) then
		form.reinit(1)
	elseif (key == KEY_2 and formID ~= 2) then
		form.reinit(2)
	elseif (key == KEY_3 and formID ~= 3) then
		form.reinit(3)
	elseif (key == KEY_4 and formID ~= 4) then
		form.reinit(4)
	end
end  

local function printForm()
    
end

--------------------------------------------------------------------------------
local function init()	
	for i=1, sequencers do
		enabled[i] = system.pLoad("enabled"..i, 0) == 1
		label[i] = system.pLoad("label"..i, trans.empty)
		switch[i] = system.pLoad("switch"..i) 
		count[i] = system.pLoad("count"..i, 2)
		values[i] = system.pLoad("values"..i, {-100, 100})
		times[i] = system.pLoad("times"..i, {50, 50})
		type[i] = system.pLoad("type"..i, 1)
		redo[i] = system.pLoad("redo"..i, 0) == 1
		finish[i] = system.pLoad("finish"..i, 0) == 1
		ctrl[i] = system.pLoad("ctrl"..i)
		
		if (enabled[i] and ctrl[i]) then
			ctrl[i] = system.registerControl(ctrl[i], label[i], "AQ"..i, "") 
			
			if (not ctrl[i]) then
				error("unable to register control")
			end
		end
	end
	
	system.registerForm(1,MENU_ADVANCED,trans.name,initForm,keyPressed,printForm)
end

--------------------------------------------------------------------------------
local function setSequence(i)
	local val = system.getInputsVal(switch[i])
		
	if (val and enabled[i] and ctrl[i] and count[i] and values[i] and times[i]) then
		if (val == 1 or (finish[i] or type[i] == 1) and busy[i]) then
			local now = system.getTimeCounter()
			
			if (val ~= 1 and type[i] == 1 and busy[i] and startTime[i] and startIndex[i]) then
				if (not back[i]) then
					-- print("Rücklauf")
					
					local offset = back[i] and -1 or 1
					local endIndex = startIndex[i]
					startIndex[i] = startIndex[i] + offset
					
					back[i] = not back[i]
					offset = back[i] and -1 or 1
					
					local duration = times[i][back[i] and endIndex or startIndex[i]] * 100
					local percent = (now - startTime[i]) / duration	
					startTime[i] = now - (1 - percent) * duration
					
					-- noch kein Rücklauf, wenn Sequenz bereits beendet war
				end
			end
			
			if (not startTime[i] or not startIndex[i]) then
				startTime[i] = now
				startIndex[i] = 1
				back[i] = nil
				busy[i] = true
			end
			
			local offset = back[i] and -1 or 1
			local endIndex = startIndex[i] + offset
			local duration = times[i][back[i] and endIndex or startIndex[i]] * 100		
			
			if (now - startTime[i] >= duration) then
				if (back[i] and endIndex == 1 or not back[i] and endIndex == count[i]) then
					if (not redo[i] and type[i] == 1 and finish[i]) then
						if (back[i]) then
							busy[i] = nil
							return
						else 
							if (val == 1) then
								local endValue = values[i][endIndex]
								system.setControl(ctrl[i], endValue / 100, 0)
								return
							end
						
							startIndex[i] = endIndex + offset
							back[i] = not back[i]
						end
					else			
						if (not redo[i] or type[i] == 2 and finish[i] and val ~= 1) then
							busy[i] = nil
							local endValue = values[i][endIndex]
							system.setControl(ctrl[i], endValue / 100, 0)
							return
						end
					
						if (redo[i]) then
							if (type[i] == 1) then
								startIndex[i] = endIndex + offset
								back[i] = not back[i]
							elseif (type[i] == 2) then
								startIndex[i] = 0
							end
						end
					end
				end
				
				offset = back[i] and -1 or 1
				startTime[i] = now
				startIndex[i] = startIndex[i] + offset
				endIndex = startIndex[i] + offset
				duration = times[i][back[i] and endIndex or startIndex[i]] * 100		
			end
			
			local startValue = values[i][startIndex[i]]
			local endValue = values[i][endIndex]
			local percent = (now - startTime[i]) / duration					
			local range = endValue - startValue
			local value = startValue + percent * range
		
			system.setControl(ctrl[i], value / 100, 0)
		else			
			resetSequence(i)
			
			local value = (type[i] == 2 or finish[i] and not (not redo[i] and type[i] == 1 and finish[i])) and values[i][count[i]] or values[i][1]
			system.setControl(ctrl[i], value / 100, 0)
		end
	end
end

--------------------------------------------------------------------------------
local function loop()
	for i=1, sequencers do
		setSequence(i)
	end
end

--------------------------------------------------------------------------------
setLanguage()
return {init=init, loop=loop, author="Thorn", version=version,name=trans.name}